/* GETUTENT ROUTINES FOR SYSTEMS THAT HAVEN'T GOT THEM */

#ifdef NOGETUTENT

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "wrttmp.h"
#include "lib_utmp.h"

int ustream;		/* file descriptor for utmp file */
int utmp_state= 0;	/* 0 = not open; 1 = open; -1 = cannot open */


/* UTMPNAME - Set the name of the utmp file.  This routine is built in on
 * most newer Unixes.  Unlike the real thing, we actually open the file here.
 * The Linux version of this returns void, while the Solaris version returns
 * an int error code.  We'd like to be able to return an error code, but for
 * Linux compatibility, we don't.
 */

void utmpname(const char *file)
{
    switch (utmp_state)
    {
    case 1:
    	close(ustream);
    	/* Drop through */
    case 0:
	if ((ustream= open(file,O_RDONLY)) < 0)
	{
	    utmp_state= 0;
	    fprintf(stderr,"Cannot open %s to read\n",file);
	    /* Unfortunately, there is no way to return an error code from
	     * this, since the standard function doesn't return an error code.
	     */
	}
	else
	{
	    utmp_state= 1;
	    fcntl(ustream,F_SETFD,1);		/* Close over execs */
	}
	break;

    default:
    	break;
    }
}


/* ENDUTENT - Close the utmp file.  This routine is built in on most
 * newer Unixes.
 */

void endutent()
{
    if (utmp_state == 1)
    {
    	close(ustream);
    	utmp_state= 0;
    }
}


/* SETUTENT - Open/Rewind the utmp file.  This routine is built in on most
 * newer Unixes.
 */

void setutent()
{
    if (utmp_state == 1)
    {
    	/* Just rewind */
    	lseek(ustream,0L,0);
    }
    else
    	utmpname(UTMP);
}


/* GETUTENT - Read the next entry from the utmp file into static storage.
 * This routine is built in on most newer Unixes.
 */

struct utmp *getutent()
{
static struct utmp ut;

    switch (utmp_state)
    {
    case 0:
	setutent();
    	/* Drop through */
    case 1:
	if (read(ustream, &ut, sizeof(struct utmp)) == sizeof(struct utmp))
	    return &ut;
    	/* Drop through */
    default:
    	return (struct utmp *)NULL;
    }
}


/* GETUTLINE - Return utmp entry for the next entry in the utmp file whose
 * ut_line field matches in->ut_line.  It's obnoxious that this wants a
 * whole utmp structure passed in instead of just a char pointer, but again,
 * we conform with Linux and Solaris.
 */

struct utmp *getutline(struct utmp *in)
{
static struct utmp ut;

    switch (utmp_state)
    {
    case 0:
	setutent();
    	/* Drop through */
    case 1:
	while (read(ustream, &ut, sizeof(struct utmp)) == sizeof(struct utmp))
	{
	    if (
#if defined(USER_PROCESS) && defined(LOGIN_PROCESS)
	        (ut.ut_type == USER_PROCESS || ut.ut_type == LOGIN_PROCESS) &&
#endif
		!strncmp(ut.ut_line, in->ut_line, UT_LINESIZE))
	    {
		return &ut;
	    }
    	}
    	/* Drop through */
    default:
    	return (struct utmp *)NULL;
    }
}

#endif /* NOGETUTENT */
