/*           MESG command for suid WRITE command
 *
 * This 'mesg' command is syntatically identical to the standard 'mesg'
 * command, but is meant for use with the suid 'write' program.  If you are
 * using the TTYPERMS option, then it is pretty much identical and you don't
 * really need it.  It does however have the advantage of patching up your
 * wrttab entry if it wasn't reset by your login (though that's not really
 * necessary), and it does support the DISCONNECT and HELPERS option and it
 * does let you set modes and preferences.
 *
 *	mesg [-s] [y|n[e]] [Y] [N] [-h[y|n|Y]] [-m[l|c|a]] [-p[w|t|a]]
 *           [-r[y|n]] [-b[y|n]]
 *
 * Without the TTYPERMS option, this changes the permission flags in the
 * /etc/wrttmp file instead of changing the write permissions on the user's tty
 * device.  It always turns the device permissions off as side effect, unless
 * you do "mesg Y".
 *
 * "mesg Y" makes your tty writable to all.  Probably there is no reason you
 * would want to do this.
 *
 * If compiled with the DISCONNECT option and "mesg N" looks for all write
 * processes who are writing you, and sends SIGTERM to them after turning off
 * your message permissions.
 *
 * The -h, -m, -p and -r flags let you set other switches besides the write
 * permissions.
 *
 * Normally the permission status after the command is executed is returned.
 * If -h, -m or -p were given without an argument, the status of that setting
 * is returned.  Otherwise, the status of the setting mentioned in the
 * argument list is returned.  -1 is returned on error.  Otherwise the codes
 * returned are
 *                   0        1       2      3
 *    ----------------------------------------
 *    messages       y        n       ye    ne
 *    helper         y        n       Y
 *    record         y        n       a
 *    preference     w        t       a     
 *    mode           l        c       a
 *    bell           y        n
 *
 * These numbers can be substituted for the usual y/n/w/t/a/l/c/a in the arg
 * lists.  This is handy for shell scripts who want to save and reset a mode.
 *
 * If the -s flag is included, all output is suppressed and the status before
 * the command was executed is returned.
 *
 * (C) Copyright, Sept 1985 - Jan Wolter
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "wrttmp.h"
#include "lib_common.h"


#ifndef ANSWER_TEL
#define ANSWER_TEL 240  /* seconds after tel during which writes are OK */
#endif

#define TRUE	1
#define FALSE	0

char mydevname[UT_LINESIZE+7];	/* my tty name in /dev/tty?? format */
char *mytty;			/* my tty name in tty?? format */
int wstream;			/* stream for the wrttmp file */
long mypos;			/* offset of my entry in wrttmp file */
struct wrthdr wt_head;		/* wrttmp file header */
struct wrttmp mywrt;		/* my wrttmp entry */
struct utmp myutmp;		/* A tmp buffer for reading utmp entries */
char *progname;

char silent= FALSE;		/* generates no output if true */
int verbose= FALSE;		/* generate whole table of output if true */

#define SMESG 0
#define SHELP 1
#define SRECO 2
#define SPREF 3
#define SMODE 4
#define SBELL 5
#define SEXCP 6		/* This must be last flag */
#define S_NUM 7

char *name[S_NUM]= {"mesg", "helper", "record", "preference",
                    "mode", "bells", "exceptions"};
char *vname[S_NUM]= {"messages", "helper flag", "record",
		     "preference (write/tel)", "mode (line/character)",
		     "bells", "exceptions"};

/* Legal tag values for each type of setting.  These must be listed in the
 * order of the corresponding numerical codes. nmap lists all legal flags,
 * smap lists only the flags that may be set by "mesg" and must have only
 * truncated versions of the nmap strings. */
char *nmap[S_NUM]= {"yn","ynY","yna","wta","lca","yn","yn"};
char *smap[S_NUM]= {"yn","ynY","yn","wta","lca","yn","yn"};

char *vmap[S_NUM][4]= { {"yes","no","yes (exceptions)", "no (exceptions)"},
			{"yes","no","always (Y)",NULL},
			{"yes","no","all",NULL},
			{"write","tel","any",NULL},
			{"line","character","any",NULL},
			{"yes","no",NULL,NULL},
			{"yes","no",NULL,NULL}};

int report= SMESG;		/* which of the above to put in return code */

char *ttyname(), *index();

void set_perms(char new_mesg, int open_up);
int report_status(void);
int get_status(int rep);
int find_me(void);
#ifdef HELPERS
int may_help(void);
#endif
#ifdef DISCONNECT
void do_disconnect(void);
#endif
#ifdef WRTHIST
int window_warning(int newmode);
#endif
#ifdef EXCEPTIONS
char *myhomedir(void);
int maywriteme(char *user, int mode);
#endif

int main(int argc, char **argv)
{
int i,j;
char new[S_NUM] = {'x', 'x', 'x', 'x', 'x', 'x', 'x'};
char disconnect= FALSE;
char open_up = FALSE;
int setting= SMESG;
char wassilent= FALSE;
int lastset= -1;
char flag;

    /* Check if I have a valid tty */
    if (!(fcntl(2,F_GETFL,0) & O_RDWR) || !isatty(2))
    {
	printf("%s: I/O stream 2 improperly redirected\n",argv[0]);
	exit(2);
    }

    /* Find my tty line */
    strcpy(mydevname,ttyname(2));
    mytty= mydevname+5;
    progname= argv[0];

    /* Parse options */
    for(i=1;i<argc;i++)
        for (j= 0; argv[i][j] != '\0'; j++)
	{
	    /* Convert number arguments to appropriate characters */
	    if (argv[i][j] >= '0' && argv[i][j] < '0'+strlen(smap[setting]))
		flag= smap[setting][argv[i][j]-'0'];
	    else
		flag= argv[i][j];

	    switch (flag)
	    {
	    case '-':
		if (j != 0) goto usage;
		break;

	    case 'v':
		verbose= TRUE;
		break;

	    case 's':
		if (setting != SMESG) goto usage;
		wassilent= silent= TRUE;
		break;

	    case 'm':
		if (setting != SMESG) report= setting;
		setting= SMODE;
		break;

	    case 'b':
		if (setting != SMESG) report= setting;
		setting= SBELL;
		break;

	    case 'r':
		if (setting != SMESG) report= setting;
		setting= SRECO;
		break;

	    case 'p':
		if (setting != SMESG) report= setting;
		setting= SPREF;
		break;
#ifdef HELPERS
	    case 'h':
		if (setting != SMESG) report= setting;
		setting= SHELP;
		break;
#endif /* HELPERS */

#ifdef EXCEPTIONS
	    case 'E':
	    case 'e':
		if (lastset != SMESG) goto usage;
		new[SEXCP]= 'y';
		break;
#endif /* EXCEPTIONS */

	    case 'Y':
		if (setting == SMESG)
		{
		    open_up= TRUE;
		    new[SMESG]= 'y';
		    lastset= SMESG;
		    break;
		}
		else if (setting != SHELP)
		    goto usage;
		/* else drop through */

	    case 'y':
	    case 'n':
	    case 'l':
	    case 'c':
	    case 'w':
	    case 't':
	    case 'a':
		if (index(smap[setting],flag) == NULL) goto usage;
		new[setting]= flag;
		lastset= setting;
		setting= SMESG;
		break;

	    case 'N':
		lastset= SMESG;
	    case 'd':
		if (setting != SMESG) goto usage;
		disconnect= TRUE;
		if (flag == 'N') new[SMESG]= 'n';
		break;

	    default:
		goto usage;
	    }
	}

#ifdef EXCEPTIONS
    /* If we changed message perms, but didn't set exceptions, turn them off */
    if (new[SMESG] != 'x' && new[SEXCP] == 'x') new[SEXCP]= 'n';
#endif /* EXCEPTIONS */

    if (setting != SMESG)
	report= setting;
    else if (report == SMESG)
	report= ((lastset > SMESG) ? lastset : SMESG);

    /* Open the utmp file */
    open_utmp();

    /* Open the wrttmp file and find us in it */
    if (init_wstream(WRTTMP,O_RDWR))
	exit(-1);
    find_me();

#ifdef HELPERS
    if ((new[SHELP] == 'y'  || new[SHELP] == 'Y') && !may_help())
    {
	printf("You are not on the helper list.\n");
	exit(1);
    }
#endif /*HELPERS*/

    /* Update permissions/helperness/preferences/bells */
    if (new[SMESG] != 'x')
	set_perms(new[SMESG],open_up);
#ifdef HELPERS
    if (new[SHELP] != 'x')
	mywrt.wrt_help= new[SHELP];
#endif /*HELPERS*/
#ifdef EXCEPTIONS
    if (new[SEXCP] != 'x')
	mywrt.wrt_except= new[SEXCP];
#endif /*EXCEPTIONS*/
    if (new[SRECO] != 'x')
	mywrt.wrt_record= new[SRECO];
    if (new[SMODE] != 'x')
	mywrt.wrt_modepref= new[SMODE];
    if (new[SPREF] != 'x')
	mywrt.wrt_telpref= new[SPREF];
    if (new[SBELL] != 'x')
	mywrt.wrt_bells= new[SBELL];

    /* Write out new settings */
    if (strncmp(new,"xxxxxxx",S_NUM))
    {
	lseek(wstream,mypos,0);
	write(wstream,&mywrt,sizeof(struct wrttmp));
	silent= TRUE;
    }

    /* Disconnect existing connections */
#ifdef DISCONNECT
    if (disconnect)
    {
	do_disconnect();
	silent= TRUE;
    }
#endif /*DISCONNECT*/

    /* Close the utmp file */
    endutent();

#ifdef WRTHIST
   if (!wassilent && (new[SMESG] == 'n' || new[SEXCP] == 'y'))
       window_warning((new[SMESG]=='n')+2*(new[SEXCP]=='y'));
#endif

    exit(report_status());

usage:
    /* Should fix up formatting of this one day */
    printf("usage: %s [-sv] ", progname);
#ifdef EXCEPTIONS
    printf("[y|n|ye|ne|Y|N|NE]");
#else
    printf("[y|n|Y|N]");
#endif
#ifdef HELPERS
    printf(" [-h[y|n|Y]]");
#endif /*HELPERS*/
    printf(" [-r[y|n]] [-m[l|c|a]] [-p[w|t|a]] [-b[y|n]]\n");
    exit(-1);
}


/* SET_PERMS:  This sets the message permissions to "new_mesg" (which is
 * either 'y' or 'n'.  It throws the tty permissions wide open as well if
 * "open_up" is true.
 */

void set_perms(char new_mesg, int open_up)
{
int mode;

    /* Change the device permissions */
    if (open_up)
	mode= PERMS_OPEN;
#ifdef TTYPERMS
    else if (new_mesg == 'y')
	mode= PERMS_ON;
#endif /*TTYPERMS*/
    else
	mode= PERMS_OFF;

    chmod(mydevname,mode);

    /* Change the wrttmp permissions */
#ifndef TTYPERMS
    mywrt.wrt_mesg= new_mesg;
#endif /*TTYPERMS*/
}


/* REPORT_STATUS -- This prints the current message permission settings and
 * returns the status code (0 = yes; 1 = no).  If silent is true, it omits
 * the printing.  If report is set to something other than SMESG it reports
 * on that other setting.  If verbose is true, it reports on all settings.
 */

int report_status()
{
int code,rc,i;

    if (verbose)
    {
	for (i= 0; i < S_NUM; i++)
	{
#ifndef HELPERS
		if (i == SHELP) continue;
#endif /*HELPERS*/
		if (i == SEXCP) continue;
		code= get_status(i);
		printf("%s: %s\n",vname[i],
		    vmap[i][code] != NULL ? vmap[i][code] : "strange");
		if (i == report) rc= code;
	}
    }
    else
    {
	rc= get_status(report);
	if (!silent)
#ifdef EXCEPTIONS
	    if (report == SMESG)
		printf("%s is %c%s\n",name[report],nmap[report][rc%2],
		       rc > 1 ? "e" : "");
	    else
#endif /*EXCEPTIONS*/
		printf("%s is %c\n",name[report],nmap[report][rc]);
    }
    return rc;
}


/* GET_STATUS -- Get current index code of given switch */

int get_status(int rep)
{
struct stat stb;
int code;

    switch (rep)
    {
    case SMESG:
#ifdef TTYPERMS
	/* Stat my tty */
	if (stat(mydevname,&stb))
	{
	    printf("%s: Error - can't stat %s\n",progname,mydevname);
	    exit(-1);
	}
	code= ((stb.st_mode & 0020)==0);
#else
	code= index(nmap[SMESG],mywrt.wrt_mesg) - nmap[SMESG];
#endif /* TTYPERMS */
#ifndef EXCEPTIONS
	break;
#else
	code+= 2 - 2*(index(nmap[SEXCP],mywrt.wrt_except) - nmap[SEXCP]);
	break;

    case SEXCP:
	code= index(nmap[SEXCP],mywrt.wrt_except) - nmap[SEXCP];
	break;
#endif /*EXCEPTIONS*/

#ifdef HELPERS
    case SHELP:
	code= index(nmap[SHELP],mywrt.wrt_help) - nmap[SHELP];
	break;
#endif /*HELPERS*/

    case SRECO:
	code= index(nmap[SRECO],mywrt.wrt_record) - nmap[SRECO];
	break;

    case SMODE:
	code= index(nmap[SMODE],mywrt.wrt_modepref) - nmap[SMODE];
	break;

    case SPREF:
	code= index(nmap[SPREF],mywrt.wrt_telpref) - nmap[SPREF];
	break;

    case SBELL:
	code= index(nmap[SBELL],mywrt.wrt_bells) - nmap[SBELL];
	break;
    }
    return code;
}


/* DO_DISCONNECT -- This does the disconnect, by searching the wrttmp file for
 * people who are writing you and sending a SIGTERM to each of the process
 * numbers listed there.  This whole approach is a bit shady.
 */

#ifdef DISCONNECT
void do_disconnect()
{
struct utmp *ut;		/* A tmp buffer for reading utmp entries */
struct wrttmp hiswrt;		/* Someone's wrttmp entry */
int slot= 0;

    /* Rewind utmp file */
    setutent();

    /* For each user who is writing me */
    for (;;)
    {
	lseek(wstream, wrttmp_offset(slot++), 0);
	if (read(wstream, &hiswrt, sizeof(struct wrttmp)) !=
	       sizeof(struct wrttmp))
	   break;

	if (!strncmp(hiswrt.wrt_what, myutmp.ut_name, UT_NAMESIZE))
	{
	    setutent();
	    /* Check apparant writer against utmp file */
	    while ((ut= getutent()) != NULL)
		if (!strncmp(hiswrt.wrt_line, ut->ut_line, UT_LINESIZE))
		{
		    /* Writer is for real: bonk him one */
		    kill(hiswrt.wrt_pid, SIGTERM);
		    break;
		}
	}
    }
}
#endif /* DISCONNECT */



/* FIND_ME -- This finds my wrttmp entry and loads it into "mywrt".
 */

int find_me()
{
struct utmp *ut;

    /* Find our entry in the Utmp file */
    if ((ut= find_utmp(mytty)) == NULL || ut->ut_name[0] == '\0')
    {
	printf("%s: Unable to find your tty (%s) in utmp file\n",
		progname,mytty);
	exit(-1);
    }
    myutmp= *ut;

    /* Find the entry in the wrttmp file */
    find_wrttmp(mytty,myutmp.ut_time,&mywrt,&mypos);
}


#ifdef HELPERS

/* MAY_HELP -- Check if the user's name is anywhere to be found in the
 * HELPERS file.  If he is there, or the file doesn't exist at all, then he
 * may designate himself as a helper.
 */

int may_help()
{
#define BUFSZ 80
FILE *hfp;
char buf[BUFSZ+1];
char myname[UT_NAMESIZE+2];

	if ((hfp= fopen(HELPERS,"r")) == NULL)
		return TRUE;
	
	strncpy(myname,myutmp.ut_name,UT_NAMESIZE);
	myname[UT_NAMESIZE]= '\0';
	strcat(myname,"\n");
	
	while (fgets(buf,BUFSZ,hfp) != NULL)
		if (!strcmp(buf,myname))
		{
			fclose(hfp);
			return TRUE;
		}

	fclose(hfp);
	return FALSE;
}
#endif /*HELPERS*/


/* WINDOW_WARNING - If there are people who can still write you under the
 * four minute window rule, print a warning message.
 */

#ifdef WRTHIST
int window_warning(int newmode)
{
struct wrthist *hist;
struct wrttmp w;
struct utmp *u;
long writer, writee;
time_t now;
int n, foundsome= 0;
FILE *fp;

    if (newmode == 0) return;

    /* Compute my slot number from my offset */
    if ((writer= (mypos - wt_head.hdr_size)/wt_head.tmp_size) > MAXSLOTS)
	return;

    if ((fp= fopen(WRTHIST,"r")) == NULL)
	return;

    fseek(fp,whindex(writer,0),0);
    hist= (struct wrthist *)malloc(MAXSLOTS*sizeof(struct wrthist));
    n= fread(hist,sizeof(struct wrthist),MAXSLOTS,fp);
    fclose(fp);

    now= time((time_t *)0);

    for (writee= 0; writee < n; writee++)
    {
	if (hist[writee].tm > myutmp.ut_time &&
	    now - hist[writee].tm <= ANSWER_TEL)
	{
	    /* Fetch "his" wrttmp entry - it may actually belong to a previous
	     * user on that line, so only the wrt_line field is really to be
	     * trusted.
	     */
	    lseek(wstream,wrttmp_offset(writee),0);
	    if (read(wstream, &w, sizeof(struct wrttmp)) !=
		   sizeof(struct wrttmp))
		continue;

	    /* Check times to ensure this is not from a later user */
	    if (hist[writee].tm < w.wrt_time)
	    	continue;

	    /* Fetch his utmp entry, and confirm that the current user was
	     * already logged in there when we sent our last telegram there.
	     */
	    if ((u= find_utmp(w.wrt_line)) == NULL || u->ut_name[0] == '\0' ||
	        hist[writee].tm < u->ut_time)
	    	continue;

#ifdef EXCEPTIONS
	    /* Check if due to exceptions he may write us anyway */
	    if (newmode > 1 && maywriteme(u->ut_name, newmode))
	    	continue;
#endif
	    if (!foundsome)
	    {
	        printf("Warning: Users to whom you recently sent messages may\n");
	        printf("still be able to write you for a few more minutes:\n");
	        foundsome= 1;
	    }
	    printf("  %-*.*s %-*.*s %4.1f more minutes\n",
	    	UT_NAMESIZE, UT_NAMESIZE, u->ut_name,
	    	UT_LINESIZE, UT_LINESIZE, u->ut_line,
	        (float)(ANSWER_TEL - now + hist[writee].tm)/60.0);
	}
    }
}


#ifdef EXCEPTIONS

/* MYHOMEDIR - Return my home directory.  Speed counts over reliability here.
 * Return NULL if we can't find it.
 */

char *myhomedir()
{
char myname[UT_NAMESIZE+2];
struct passwd *pw;
char *dir, *getenv();

    /* Try environment variable first */
    if ((dir= getenv("HOME")) != NULL)
	return dir;

    /* If that don't work, try passwd file */
    strncpy(myname,myutmp.ut_name,UT_NAMESIZE);
    myname[UT_NAMESIZE]= '\0';
    if ((pw= getpwnam(myname)) != NULL)
    	return pw->pw_dir;

    /* If that don't work, give it up */
    return NULL;
}


/* MAYWRITEME - can the named user write me?  This checks exceptions.  Does
 * some caching so that multiple successive calls are efficient.
 */

int maywriteme(char *user, int mode)
{
static int oldmode= -1;
static char *list;
struct stat st;
char *dir,*fname, *p,*q;
FILE *fp;
int n, l, nl, sp;

    if (mode < 2)
	return (mode == 0);

    if (mode != oldmode)
    {
	list= NULL;
	oldmode= mode;

	/* Open the appropriate exception file */
	if ((dir= myhomedir()) == NULL)
	    return FALSE;
	fname= (char *)malloc(strlen(dir)+15);
	sprintf(fname,"%s/%s",dir, mode==2 ? ".nowrite" : ".yeswrite");
	fp= fopen(fname,"r");
	free(fname);

	/* Load the file into memory */
	if (fp != NULL && !fstat(fileno(fp),&st))
	{
	    list= (char *)malloc(st.st_size+1);
	    n= fread(list,1,st.st_size,fp);
	    list[n]= '\0';
	}

	if (list != NULL)
	{
	    /* Compress list into login names separated by single spaces */
	    nl= 1; sp= 1;
	    for (p= q= list; *p != '\0'; p++)
	    {
		/* Skip lines starting with # */
		if (nl && (*p == '#'))
		{
		    p= strchr(p,'\n');
		    if (p == NULL) break;
		}

		nl= (*p == '\n');

		if (strchr(" ,;:\t\n\r",*p))
		{
		    if (!sp) *(q++)= ' ';
		    sp= 1;
		}
		else
		{
		    sp= 0;
		    *(q++)= *p;
		}
	    }
	    *q= '\0';
	}
    }

    if (list != NULL)
    {
	/* Find length of user name */
	for (l= 0; l < UT_NAMESIZE && user[l] != '\0'; l++)
	    ;

	/* Search compressed list for the name */
	p= list;
	for (;;)
	{
	    if (!strncmp(p,user,l) && (p[l] == ' ' || p[l] == '\0'))
	    	return (mode != 2);
	    if ((p= strchr(p,' ')) == NULL) break;
	    p++;
	}
    }
   
    /* Not on list */
    return (mode == 2);
}
#endif /*EXCEPTIONS*/
#endif /*WRTHIST*/
